% Statistical and Numerical Methods for Chemical Engineers 2023
% Exercise 4, Assignment 1
clear all; close all; clc

% Define ODE system with initial conditions
k=1;
f= @(t,y) -k.*y;
y0=linspace(0,1,20);
t=linspace(0,10);

hfig=vector_field(f,y0,t);

% The solver finds a function, which evaluated at each disrete point,
% follows the characteristic behavior defined for its derivative. This
% characteristic behavior is the ODE, which follows the rule dy/dt =
% f(t,y). The solutions are unique, this means that the trajectories do not
% intersect.

% 1.2 Solve the problem with the forward (explicit) euler method
t0 = 0;
tEnd = 10;
y0 = 10;
h = 0.1;
forward_stepper = @(k,y,h) y - h*k*y;
[T,Y] = stepper_integrate(forward_stepper,k,t0,tEnd,y0,h);

% 1.3 Solve the problem with the backward (implicit) euler method
backward_stepper = @(k,y,h) y/(1 + h*k);
[T2,Y2] = stepper_integrate(backward_stepper,k,t0,tEnd,y0,h);

% 1.4 Plot the solutions
figure
subplot(2,2,1)
t2 = t0:h:tEnd;
plot(t2,y0*exp(-k*t2)) % analytical solution forward euler
hold on
plot(T,Y,'ro') % numeric solution
hold off
title('Forward')
xlabel('Time')
ylabel('Concentration')
subplot(2,2,2)
plot(t2,y0*exp(-k*t2)) % analytical solution 
hold on
plot(T2,Y2,'ro') % numeric solution backward euler
hold off
title('Backward')
xlabel('Time')
ylabel('Concentration')
% The numeric solutions diverge from the analytical with increasing h,
% i.e. the error increases with increasing h. For h above 2 the Forward
% Euler Method clearly becomes instable.

% 1.5 Plot the global error
H = logspace(-3,0.5,10);
Er_f = zeros(1,length(H)); % vectorsd storing global error
Er_b = zeros(1,length(H));
y_an = y0*exp(-k*tEnd);
for i = 1:length(H)
    [T,Y] = stepper_integrate(forward_stepper,k,t0,tEnd,y0,H(i));
    [T2,Y2] = stepper_integrate(backward_stepper,k,t0,tEnd,y0,H(i));
    Er_f(i) = abs(y_an-Y(end));
    Er_b(i) = abs(y_an-Y2(end));
end
subplot(2,2,3)
loglog(H,Er_f,'rx')
hold on
loglog(H,Er_b,'ko')
legend('Forward','Backward','location','Best')
title('Global Error Comparison')
xlabel('h')
ylabel('global error')
hold off
% You can observe the onset of the instability of the Forward Method at h =
% 2e0.

  function [t,y] = stepper_integrate(stepper,k,t0,tEnd,y0,h)
  % Determins the value of the integrated ODE function y for each
  % time step t.
  
  % Outputs:
  % time step vector t
  % concentration vector y
  
  % Inputs:
  % rules for integration at each step: stepper (function handle)
  % decay rate k (scalar)
  % start time t0 (scalar)
  % end time tEnd (scalar)
  % initial concentration (scalar)
  % step size h
  
  N = round((tEnd-t0)/h);
  t = zeros(1,N+1);
  y = zeros(1,N+1);
  t(1) = t0;
  y(1) = y0;
  for j=1:N
    t(j+1) = t(j) + h;
    y(j+1) = stepper(k,y(j),h);
  end
  end
